"use strict";

let Url = require('url-parse');
let beanStalk = require('nodestalker');
let Aws = require('aws-sdk');
let Fs = require('fs');

class PractoQueue{
  static getInstance(uri, sqsSpoolDir=null){
    let parts = new Url(uri);
    parts.protocol = parts.protocol.replace(':','');
    if(parts.protocol === 'https'){
      const host = parts.hostname;
      const hostParts = host.split('.');
      if((hostParts[0] === 'sqs') && hostParts[2] === 'amazonaws' && hostParts[3] === 'com'){
        return new Sqs(uri, sqsSpoolDir);
      }
      else{
        throw 'Invalid Sqs URI';
      }
    }
    else if(parts.protocol === 'beanstalk'){
      return new Beanstalk(uri);
    }
    else{
      throw 'Unsupported URI scheme';
    }
  }
}

class Queue{
   getQueueName(queueIdentifier, host=null){
    let queueName = queueIdentifier;
    if(host){
      const baseUrl = host.split('/')[2];
      const subDomain = baseUrl.split('.')[0];
      const subDomainParts = subDomain.split('-');
      if(subDomainParts.length > 1){
        queueName = queueName + '-' + subDomainParts[1]
      }
    }
    return queueName;
  }

  setDefaultHost(host=null){
    this.host = host;
  }
}

class Sqs extends Queue{
  constructor(uri, sqsSpoolDir=null){
    super();
    this.sqsSpoolDir = sqsSpoolDir;
    let parts = new Url(uri);
    let host = parts.hostname;
    let hostParts = host.split('.', 2);
    this.region = hostParts[1];
    const awsAccessKeyId = parts.username;
    const awsSecretAccessKey = parts.password;
    Aws.config = new Aws.Config();
    Aws.config.accessKeyId = awsAccessKeyId;
    Aws.config.secretAccessKey = awsSecretAccessKey;
    Aws.config.region = this.region;
    this.sqs = new Aws.SQS();
    let path = parts.pathname;
    if(path.startsWith('/')){
      path = path.substr(1);
    }
    let pathParts = path.split('/');
    this.accountId = pathParts[0];
    parts.protocol = parts.protocol.replace(':','');
    this.protocol = parts.protocol;
    this.queuePrefix = '';
    let lastPart = pathParts[1];
    if((lastPart.endsWith('_')) || (lastPart.endsWith('-'))) {
      this.queuePrefix = lastPart;
    }
    this.queue = null;
    this.queueName = '';
    this.host = null;
    this.BASE64_EXPAND_RATIO = 1.37;
    this.MAX_JOB_SIZE = 262144;
  }

  setDefaultQueue(queueName, callback){
    queueName = this.getQueueName(queueName, this.host);
    this.queueName = this.queuePrefix + queueName;
    this.sqs.getQueueUrl({QueueName: this.queueName}, (err, data) =>{
      if (err) {
        if(err.statusCode == 400){
          let params = {
            QueueName: queueName
          }
          this.sqs.createQueue(params, function(err, data) {
            if (err){
              callback(err);
            }
            else {
              callback(null, data.QueueUrl);
            }
          });
        }else{
          callback(err);
        }
      } else {
        callback(null, data.QueueUrl);
      }
    });
  }

  insertJob(jobBody, queueName, delay=0){
    jobBody = new Buffer(jobBody).toString('base64');
    if(delay > 900){
      delay = 900;
    }
    let sqsParams = {
      MessageBody: jobBody,
      QueueUrl: this.queue,
      DelaySeconds: delay
    };
    let queue = this.queue;
    if(this.sqsSpoolDir){
      // LOGIC HERE FOR SPOOL
      let jobDir = '';
      queueName = this.queuePrefix + queueName;
      queueName = this.getQueueName(queueName, this.host);
      if (!Fs.existsSync(this.sqsSpoolDir)){
        Fs.mkdirSync(this.sqsSpoolDir);
      }
      if(!jobBody){
        throw 'Job body cannot be empty';
      }
      if(jobBody.length * this.BASE64_EXPAND_RATIO > this.MAX_JOB_SIZE){
        throw 'One or more parameters are invalid. Reason:' +
              ' Message must be shorter than 262144 bytes.';
      }
      let currentTime = (+new Date)/1000;
      jobDir = parseInt(currentTime/100).toString();
      if (!Fs.existsSync(this.sqsSpoolDir + '/' + jobDir)){
        try {
          Fs.mkdirSync(this.sqsSpoolDir + '/' + jobDir)
        } catch (err) {
          if (err.code !== 'EEXIST') throw err
        }
      }

      currentTime = +new Date;
      let fileName = parseInt(currentTime).toString() + this.getRandomString(10);
      const encodeBody = new Buffer(jobBody).toString('base64');
      const data = [this.region,
                    this.getSqsQueueUrl(queueName),
                    encodeBody,
                    delay.toString(),
                    '60'
                   ]
      const payload = data.join('\t');
      let filePath = this.sqsSpoolDir + '/' + jobDir + '/' + fileName;
      Fs.writeFile(filePath, payload, function(err){
        if(err){
          throw "Error while writing file";
        }
      });
      Fs.rename(filePath, filePath + '.v1message', function(err) {
        if ( err ){
          throw 'ERROR' + err;
        }
      });
    }
    else{
      if(queueName !== this.queueName){
        queueName = this.queuePrefix + queueName;
        queueName = this.getQueueName(queueName, this.host);
        this.sqs.getQueueUrl({QueueName: queueName}, (err, data) =>{
          if (err) {
            if(err.statusCode == 400){
              let params = {
                QueueName: queueName
              }
              this.sqs.createQueue(params, (err, data)=> {
                if (err){
                  throw 'ERR: ' + err;
                }
                else {
                  sqsParams.QueueUrl = data.QueueUrl;
                  this.sqs.sendMessage(sqsParams, function(err, data) {
                    if (err) {
                      throw 'ERR: ' + err;
                    }
                  });
                }
              });
            }else{
              throw 'ERR: ' + err;
            }
          }
          else {
            sqsParams.QueueUrl = data.QueueUrl;
            this.sqs.sendMessage(sqsParams, function(err, data) {
              if (err) {
                throw 'ERR: ' + err;
              }
            });
          }
        });
      }
      else{
        this.sqs.sendMessage(sqsParams, function(err, data) {
          if (err) {
            throw 'ERR: ' + err;
          }
        });
      }
    }
  }

  getSqsQueueUrl(queueName){
    return this.protocol + "://sqs." + this.region + ".amazonaws.com/" +
           this.accountId + "/" + queueName;
  }

  deleteJob(queueJob, queueName){
    const queueUrl = this.getSqsQueueUrl(queueName);
    let sqsParams = {
      QueueUrl: queueUrl,
      ReceiptHandle: queueJob.ReceiptHandle
    };
    this.sqs.deleteMessage(sqsParams, function(err, data) {
    });
  }

  getRandomString(count){
    const base = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789_-";
    let ret = ""
    let strlen = base.length;
    for(let i=0; i<count; i++){
      ret = ret + base[Math.floor(Math.random() * (strlen - 0) + 0)];
    }
    return ret
  }

  getJob(queueName, callback){
    const queueUrl = this.getSqsQueueUrl(queueName);
    let sqsParams = {
      QueueUrl: queueUrl,
      WaitTimeSeconds: 20
    };
    this.sqs.receiveMessage(sqsParams, function(err, data) {
      while(1){
        if(data.Messages){
          break;
        }
      }
      const message = data.Messages[0];
      const messageBody = message.Body;
      const messageId = message.MessageId;
      const receiptHandle = message.ReceiptHandle;
      let queueJob = new QueueJob(messageId, messageBody);
      callback(queueJob);
    });
  }
}

class Beanstalk extends Queue{
  constructor(uri){
    super();
    let parts = new Url(uri);
    let host = parts.hostname;
    let port = 11300;
    if(parts.port){
      port = parts.port;
    }
    let path = parts.pathname;
    if(path.startsWith('/')){
      path = path.substr(1);
    }
    let pathParts = path.split('/');
    this.queuePrefix = '';
    let lastPart = pathParts[0];
    if((lastPart.endsWith('_')) || (lastPart.endsWith('-'))) {
      this.queuePrefix = lastPart;
    }
    const connectionString = host + ':' + port;
    this.beanStalkClient = beanStalk.Client(connectionString);
    this.host = null;
  }

  setDefaultQueue(queueName){
    queueName = this.getQueueName(queueName, this.host);
    this.queueName = this.queuePrefix + queueName;
  }

  insertJob(jobBody, queueName, delay=0){
    queueName = this.getQueueName(queueName, this.host);
    this.beanStalkClient.use(this.queuePrefix + queueName).onSuccess((data) =>{
      this.beanStalkClient.put(jobBody).onSuccess((data) =>{
        console.log(data);
    	  //  return data;
      });
    });
  }

  deleteJob(queueJob){
     this.beanStalkClient.deleteJob(queueJob.id).onSuccess(() => {
     });
  }

  getJob(callback){
    this.beanStalkClient.watch(this.queueName).onSuccess((data) => {
      this.beanStalkClient.reserve().onSuccess((job) => {
        let queueJob = new QueueJob(job.id, job.data);
        callback(queueJob);
      });
    });
  }
}

class QueueJob{
  constructor(jobId, body){
    this.jobId = jobId;
    this.body = body;
  }

  getBody(){
    return this.body;
  }

  body(){
    return this.body;
  }

  getId(){
    return this.id;
  }

  getReceiptHandle(){
    return this.receiptHandle;
  }

  setReceiptHandle(receiptHandle){
    return this.receiptHandle = receiptHandle
  }
}

module.exports = {
  PractoQueue: PractoQueue,
  Queue: Queue,
  Beanstalk: Beanstalk,
  QueueJob: QueueJob,
  Sqs: Sqs
}
